/* GNU POP3 - a small, fast, and efficient POP3 daemon
   Copyright (C) 1999 Jakob 'sparky' Kaivo <jkaivo@nodomainname.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <time.h>
#include "gnu-pop3d.h"

#define LINESLICE 160

/* Takes a string as input and returns either the remainder of the string
   after the first space, or a zero length string if no space */

char *
pop3_args (const char *cmd)
{
  int space = -1, i = 0, len;
  char *buf;

  len = strlen (cmd) + 1;
  buf = calloc (len, sizeof (char));
  if (buf == NULL)
	pop3_abquit (ERR_NO_MEM);

  while (space < 0 && i < len)
	{
	  if (cmd[i] == ' ')
		space = i + 1;
	  else if (cmd[i] == '\0' || cmd[i] == '\r' || cmd[i] == '\n')
		len = i;
	  i++;
	}

  if (space < 0)
	buf[0] = '\0';
  else
	{
	  for (i = space; i < len; i++)
		if (cmd[i] == '\0' || cmd[i] == '\r' || cmd[i] == '\n')
		  buf[i - space] = '\0';
		else
		  buf[i - space] = cmd[i];
	}

  return buf;
}

/* This takes a string and returns the string up to the first space or end of
   the string, whichever occurs first */

char *
pop3_cmd (const char *cmd)
{
  char *buf;
  int i = 0, len;

  len = strlen (cmd) + 1;
  buf = calloc (len, sizeof (char));
  if (buf == NULL)
	pop3_abquit (ERR_NO_MEM);

  for (i = 0; i < len; i++)
	{
	  if (cmd[i] == ' ' || cmd[i] == '\0' || cmd[i] == '\r' || cmd[i] == '\n')
		len = i;
	  else
		buf[i] = cmd[i];
	}
  buf[i - 1] = '\0';
  return buf;
}

/* Returns OK if a message actually exists and has not been marked by dele(),
   otherwise ERR_NO_MESG */

int
pop3_mesg_exist (int mesg)
{
  if ((mesg < 0) || (mesg >= num_messages) || (messages[mesg].deleted == 1))
	return ERR_NO_MESG;

  return OK;
}

/* This is called if it needs to quit without going to the UPDATE stage.
   This is used for conditions such as out of memory, a broken socket, or
   being killed on a signal */

int
pop3_abquit (int reason)
{
  pop3_unlock ();
  if (mbox != NULL)
	fclose (mbox);
  if (messages != NULL)
	free (messages);

/* Jonathan Chin <jonathan.chin@ox.compsoc.net>
   Ensure that ofile is non-null so following routine will
   not attempt to write an error message through the socket.
*/

  switch (reason)
	{
	case ERR_NO_MEM:
	  if (NULL != ofile)
		fprintf (ofile, "-ERR Out of memory, quitting\r\n");
	  syslog (LOG_ERR, "Out of memory");
	  break;
	case ERR_DEAD_SOCK:
	  ofile = NULL;
	  if (NULL != ofile)		/* should this even print to socket? */
		fprintf (ofile, "-ERR Socket closed, quitting\r\n");
	  syslog (LOG_ERR, "Socket closed");
	  break;
	case ERR_SIGNAL:
	  if (NULL != ofile)
		fprintf (ofile, "-ERR Quitting on signal\r\n");
	  break;
	case ERR_TIMEOUT:
	  if (NULL != ofile)
		fprintf (ofile, "-ERR Session timed out\r\n");
	  if (state == TRANSACTION)
		syslog (LOG_INFO, "Session timed out for user: %s", username);
	  else
		syslog (LOG_INFO, "Session timed out for no user");
	  break;
#ifdef IP_BASED_VIRTUAL
	case ERR_IP_BASED:
	  if (NULL != ofile)
		fprintf (ofile, "-ERR Quitting - host name not found\r\n");
	  syslog (LOG_ERR, "Quitting - host name not found");
	  break;
#endif /* IP_BASED_VIRTUAL */
	case ERR_NO_OFILE:
	  syslog (LOG_ERR, "Quitting - couldn't open stream");
	  break;
	default:
	  if (NULL != ofile)
		fprintf (ofile, "-ERR Quitting (reason %d unknown)\r\n", reason);
	  syslog (LOG_ERR, "Unknown quit: %d", reason);
	  break;
	}
  if (NULL != ofile)
	fflush (ofile);
  _exit (1);
}

/* This locks the mailbox file, works with procmail locking */

int
pop3_lock (void)
{
  struct flock fl;

  fl.l_type = F_WRLCK;
  fl.l_whence = SEEK_SET;
  fl.l_start = 0;
  fl.l_len = 0;
  if (fcntl (fileno (mbox), F_SETLK, &fl) == -1)
	{
	  syslog (LOG_ERR, "Failed fcntl() file lock: %s\n", strerror (errno));
	  return ERR_LOCK_FAILED;
	}
  lockfile = malloc (strlen (mailbox) + strlen (".lock") + 1);
  if (lockfile == NULL)
	pop3_abquit (ERR_NO_MEM);
  strcpy (lockfile, mailbox);
  strcat (lockfile, ".lock");
  lock = fopen (lockfile, "r");
  if (lock != NULL)
	{
	  free (lockfile);
	  lockfile = NULL;
	  return ERR_MBOX_LOCK;
	}
  lock = fopen (lockfile, "w");
  if (lock == NULL)
	{
	  syslog (LOG_ERR, "Failed to create lockfile: \"%s\"\n", lockfile);
	  free (lockfile);
	  lockfile = NULL;
	  return ERR_LOCK_FAILED;
	}
  return OK;
}

/* Unlocks the mailbox */

int
pop3_unlock (void)
{
  struct flock fl;
  if (debug & 0x40)
	syslog (LOG_INFO, "Unlocking start");
  fl.l_type = F_UNLCK;
  fl.l_whence = SEEK_CUR;
  if (mbox != NULL)
	{
	  if (debug & 0x40)
		syslog (LOG_INFO, "Unlocking mailbox");
	  fcntl (fileno (mbox), F_SETLK, &fl);
	}
  if (lockfile != NULL)
	{
	  if (debug & 0x40)
		syslog (LOG_INFO, "Unlocking file %s", lockfile);
	  fclose (lock);
	  if (unlink (lockfile) == -1)
		{
		  syslog (LOG_ERR, "Removing lockfile %s: %s\n", lockfile, strerror (errno));
		  free (lockfile);
		  lockfile = NULL;
		  return ERR_FILE;
		}
	  free (lockfile);
	  lockfile = NULL;
	}
  if (debug & 0x40)
	syslog (LOG_INFO, "Unlocking end");
  return OK;
}

/* This parses the mailbox to get the number of messages, the size of each
   message, and the read offset of the first character of each message. This
   is where the real speed magic is done */

int
pop3_getsizes (void)
{
  char buf[LINESLICE * sizeof (char)];
  unsigned int max_count = 0;
  /* start with 1 because first line of mbox is From */
  int empty_line = 1;
  message *tmp;
  unsigned int lbuf = 0, i;
  char *tmp1 = NULL, *res;
  fpos_t initfpos;

  int header = 1;
  struct md5_ctx md5context;

  num_messages = 0;

  if (messages == NULL)
	{
	  messages = calloc (11, sizeof (message));
	  if (messages == NULL)
		pop3_abquit (ERR_NO_MEM);
	  max_count = 10;
	}

/*  if (mdn) */
  fgetpos (mbox, &(messages[num_messages].msgstart));
  initfpos = messages[num_messages].msgstart;

  while ((res = fgets (buf, LINESLICE, mbox)) != NULL)
	{
	  /* Only count a message if "From " is a real mbox delimiter.
	     It depends on the delivery agent; some escape all "From " lines;
	     some don't. */
	  if (empty_line && !strncmp (buf, "From ", 5))
		{

		  empty_line = 0;
		  if ((num_messages > 0) && header)
			md5_finish_ctx (&md5context, messages[num_messages - 1].md5digest);

		  if ((debug & 0x80) && (num_messages > 0))
			syslog (LOG_ERR, "getsizes: msg= %d size= %d\n", num_messages,
					messages[num_messages - 1].size);
		  num_messages++;

		  if (num_messages > max_count)
			{
			  max_count = num_messages * 2;
			  tmp = realloc (messages, (max_count + 1) * sizeof (message));
			  if (tmp == NULL)
				pop3_abquit (ERR_NO_MEM);
			  else
				messages = tmp;
			}

		  fgetpos (mbox, &(messages[num_messages - 1].header));
		  messages[num_messages - 1].size = 0;
		  messages[num_messages - 1].deleted = 0;
		  messages[num_messages - 1].status[0] = '\0';	/* Status: RFC 2076 */
		  messages[num_messages - 1].chng = 0;	/*  msg disp notif sent/unread msg retrieved */
		  for (i = 0; i < 16; i++)
			messages[num_messages - 1].md5digest[i] = '\0';
		  header = 1;
		  md5_init_ctx (&md5context);
		  md5_process_bytes (buf, strlen (buf), &md5context);
		  while ((res != NULL) && (strchr (buf, '\n') == NULL))
			{
			  res = fgets (buf, LINESLICE, mbox);
			  if (res != NULL)
				md5_process_bytes (buf, strlen (buf), &md5context);
			}
		}
	  else
		{						/* ordinary message lines hdr + body */
		  lbuf = strlen (buf);
		  if (header && num_messages > 0)
			{
			  if ((buf[0] == '\r') || (buf[0] == '\n') || (buf[0] == '\0'))
				{
				  header = 0;
				  if (strchr (messages[num_messages - 1].status, 'O') == NULL)
					{
					  strcat (messages[num_messages - 1].status, "O");	/* Old msg */
					  messages[num_messages - 1].chng = 1;
					}
/* printf("num: %d hpos: %ld ch: %d,%s\n", num_messages, messages[num_messages - 1].header.__pos, messages[num_messages - 1].chng,
 				messages[num_messages - 1].status); */
				  md5_finish_ctx (&md5context, messages[num_messages - 1].md5digest);
				}
			  else if ((strncmp (buf, "X-", 2) != 0) && (strncmp (buf, "Status:", 7) != 0))	/* Skip X-.... and Status: lines */
				{
				  md5_process_bytes (buf, lbuf, &md5context);
				}
			  else
				{				/* extract status from and insert into table */
				  tmp1 = pop3_pickupheader (buf, "Status");
				  if (tmp1)
					{
					  strncpy (messages[num_messages - 1].status, tmp1, 7);
					  free (tmp1);
					}
				}
			}

		  if (buf[0] == '\n' || buf[0] == '\r' || buf[0] == '\0')
			empty_line = 1;
		  else if (num_messages > 0)
			empty_line = 0;
		  if (num_messages > 0)
			messages[num_messages - 1].size += lbuf;
		  res = buf;
		  while ((res != NULL) && (strchr (buf, '\n') == NULL))
			{					/* read possible remaining part of line */
			  res = fgets (buf, LINESLICE, mbox);
			  if (res != NULL)
				{
				  lbuf = strlen (buf);
				  if (header && lbuf > 0 && num_messages > 0)
					md5_process_bytes (buf, lbuf, &md5context);
				  if (num_messages > 0)
					messages[num_messages - 1].size += lbuf;
				}
			}					/* while */

		  /* RFC 1939 sect. 11 EOL must always be counted as 2 chars (CR LF) */
		  if (!strchr (buf, '\r') && num_messages > 0)
			messages[num_messages - 1].size++;
		}

	  /* avoid calling fgetpos for every line, only for potential lines before the "From " line */
	  if (empty_line == 1)
		fgetpos (mbox, &(messages[num_messages].msgstart));	/* start of next msg */

	}
  if ((num_messages > 0) && header)
	md5_finish_ctx (&md5context, messages[num_messages - 1].md5digest);
  if (memcmp (&messages[0].msgstart, &initfpos, sizeof (fpos_t)) != 0)
	syslog (LOG_ERR, "Mailbox not starting with 'From '. Skipping initial lines");
  return OK;
}

/* Prints out usage information and exits the program */

void
pop3_usage (char *argv0)
{
  printf ("Usage: %s [OPTIONS]\n", argv0);
  printf ("Runs the GNU POP3 daemon.\n\n");
  printf
	("  -D, --debug=DEBUGLEVEL   report debugging information. 2 hex digits (child, daemon). Default -D00\n");
  printf ("  -d, --daemon=MAXCHILDREN runs in daemon mode with a maximum\n");
  printf ("                           of MAXCHILDREN child processes. Default: -d10\n");
  printf ("  -h, --help               display this help and exit\n");
  printf ("  -i, --inetd              runs in inetd mode (default)\n");
  printf ("  -p, --port=PORT          specifies alternate port to listen on, implies -d\n");
  printf ("  -u, --user=USER          name or user id to use if not otherwise specified\n");
  printf ("  -g, --group=GROUP        name or group id to use if not otherwise specified\n");
  printf ("  -t, --timeout=TIMEOUT    sets idle timeout to TIMEOUT seconds\n");
  printf
	("  -m, --mdn                Message Disposition Notification (return receipts) on by default\n");
  printf ("  -X, --tmppath=TMPPATH    Directory for temporary files  Default: /tmp\n");
  printf ("  -v, --version            display version information and exit\n");
}

/* Default signal handler to call the pop3_abquit() function */
/* Do not use debug level 0x80 unless needed, as it may conflict with futex waits and hang the process */

void
pop3_signal (int sig, siginfo_t * siginfo, void *ucontext)
{
  int stat;
  int curr_exit_flag = exit_flag;
  pid_t mypid = getpid ();
  pid_t sigchild;
  char buf[128];
  sigchild = 0;
  switch (sig)
	{
	case SIGCHLD:
	  while ((sigchild = waitpid (-1, &stat, WNOHANG)) > 0)
		{
		  if (debug & 0x08)
			{
			  sprintf (buf, "PID: %d Terminated child %d", mypid, sigchild);
			  pop3_syslog (LOG_INFO, buf);
			}
		  child_procs--;
		}
	  sigchild = siginfo->si_pid;
	  break;

	case SIGHUP:
	  exit_flag = 0;
	case SIGTERM:				/* terminate gracefully */
	  exit_flag++;
	  if (debug & 0x08)
		{
		  sprintf (buf, "PID: %d Quitting gracefully on signal %s", mypid,
				   (sig == SIGHUP ? "SIGHUP" : "SIGTERM"));
		  pop3_syslog (LOG_INFO, buf);
		}
	  debug = 0x7f;
	  if (sock != -1)
		{
		  if (debug & 0x80)
			{
			  sprintf (buf, "PID: %d Closing socket", mypid);
			  pop3_syslog (LOG_INFO, buf);
			}
		  close (sock);
		  sock = -1;
		}
	  if (curr_exit_flag != 1)
		{
		  if (debug & 0x08)
			{
			  sprintf (buf, "PID: %d Terminating %d processes", mypid, child_procs);
			  pop3_syslog (LOG_INFO, buf);
			}
		  kill (0, SIGTERM);
		}
	  break;

	default:
	  if (curr_exit_flag > 1)
		{
		  if (debug & 0x08)
			{
			  sprintf (buf, "pop3_signal: PID: %d default exit_flag = %d", mypid, curr_exit_flag);
			  pop3_syslog (LOG_INFO, buf);
			}
		  break;				/* already been here */
		}
	  exit_flag = 2 | sig << 2;
	  if (debug & 0x08)
		{
		  sprintf (buf, "PID: %d Quitting on signal: %d", mypid, sig);
		  pop3_syslog (LOG_INFO, buf);
		}
	}

  if (debug & 0x08)
	{
	  sprintf (buf, "PID: %d Signal processed: %d from %d Active children: %d", mypid, sig,
			   sigchild, child_procs);
	  pop3_syslog (LOG_INFO, buf);
	}
}

/* In Linux there is a problem about calling syslog from a signal handler especiallly during a DoS attack
   or when many clients are connecting. The server or a child process will hang in a futex_wait.
   Logging is accomplished by storing the messages and using syslog when calling from outside the handler
   The problem may be caused my malloc.
*/

void
pop3_syslog (int priority, const char *msg)
{
#define MAXMSG 500
  static char *msgs[MAXMSG];
  static int msgcount = -1;
  int i;
  if (msgcount == -1)
	{							/* initialize */
	  for (i = 0; i < MAXMSG; i++)
		{
		  msgs[i] = calloc (255, sizeof (char));
		}
	  msgcount = 0;
	}

  /* FLushing the buffer when msg = NULL. */
  if (msg == NULL)
	{
	  for (i = 0; i < msgcount; i++)
		{
		  syslog (priority, "%s", msgs[i]);
		  msgs[i][0] = '\0';
		}
	  msgcount = 0;
	}
  else if (msgcount < MAXMSG)
	{
	  strncpy (msgs[msgcount++], msg, 255);
	}
}

/* Gets a line of input from the client */

char *
pop3_readline (int fd)
{
  fd_set rfds;
  struct timeval tv;
  char *buf;
  int done = 0;
  int too_long = 0;
  int p = 0;

  if ((buf = calloc (1, POP_MAXCMDLEN + 1)) == NULL)
	pop3_abquit (ERR_NO_MEM);


/* continue to read until newline is received */
  do
	{

	  FD_ZERO (&rfds);
	  FD_SET (fd, &rfds);
	  tv.tv_sec = timeout;
	  tv.tv_usec = 0;

	  if ((timeout > 0) && (0 >= select (fd + 1, &rfds, NULL, &rfds, &tv)))
		pop3_abquit (ERR_TIMEOUT);

	  if (read (fd, &buf[p], POP_MAXCMDLEN - p) <= 0)
		pop3_abquit (ERR_DEAD_SOCK);

	  p = strlen (buf);

	  if (strchr (buf, '\n'))
		done = 1;
	  else if (p >= POP_MAXCMDLEN)
		{
		  memset (buf, '\0', POP_MAXCMDLEN + 1);
		  too_long = 1;
		  p = 0;
		}
	}
  while (!done);

  /* only return text if has linefeed in first read() characters */
  if (too_long)
	buf[0] = '\0';

  if (debug & 0x80)
	syslog (LOG_ERR, "rdlin: %s\n", buf);
  return buf;
}

/* Deallocation of the password structure */

void
pop3_freepw (struct passwd *pwdentry)
{
  if (pwdentry != NULL)
	{
/*  fprintf(stderr, "pw_name %s\n", pwdentry->pw_name);
    fprintf(stderr, "pw_passwd %s\n", pwdentry->pw_passwd);
    fprintf(stderr, "pw_gecos %s\n", pwdentry->pw_gecos);
    fprintf(stderr, "pw_dir %s\n", pwdentry->pw_dir);
    fprintf(stderr, "pw_shell %s\n", pwdentry->pw_shell);
*/
	  free (pwdentry->pw_name);
	  free (pwdentry->pw_passwd);
	  free (pwdentry->pw_gecos);
	  free (pwdentry->pw_dir);
	  free (pwdentry->pw_shell);
	  free (pwdentry);
	}
  return;
}

/* If the MTA has written into old mailbox from before the renaming, recover the data from it.
   This may happen in case of a race condition.
   Called after mailbox has been locked.
 */
int
pop3_recover_oldmailbox (char *mailbox, FILE * mbox)
{
  int rc = 0;
  ssize_t lbuf, chwrt;
  char *oldmailbox = NULL;
  FILE *oldmbox;
  struct stat st;
  struct flock fl;
  char buf[4096];

  oldmailbox = malloc (sizeof (char) * (strlen (mailbox) + 14));	/* 4 is not enough to avoid signal 11 at free !! */
  if (oldmailbox == NULL)
	return ERR_NO_MEM;

  strcpy (oldmailbox, mailbox);
  strcat (oldmailbox, ".old");

  rc = stat (oldmailbox, &st);
  if (rc != 0)
	return rc;
/* open existing oldmailbox */
  oldmbox = fopen (oldmailbox, "r+");
  if (oldmbox == NULL)
	return ERR_FILE;
  rewind (oldmbox);
  if (fseek (mbox, 0L, SEEK_END))
	{
	  free (oldmailbox);
	  return ERR_FILEPOS;
	}

/* lock it */

  fl.l_type = F_WRLCK;
  fl.l_whence = SEEK_SET;
  fl.l_start = 0;
  fl.l_len = 0;
  if (fcntl (fileno (oldmbox), F_SETLK, &fl) == -1)
	{
	  syslog (LOG_ERR, "Failed fcntl() file lock:%s %s\n", oldmailbox, strerror (errno));
	  free (oldmailbox);
	  return ERR_LOCK_FAILED;
	}
/* get size of it */
  rc = fstat (fileno (oldmbox), &st);
  if (rc == 0)
	{
	  errno = 0;
	  if (st.st_size != 0)
		{
		  syslog (LOG_INFO, "%ld bytes found in old mailbox: %s", st.st_size, oldmailbox);

/* if size > 0 append contents to current mailbox */
		  while ((lbuf = read (fileno (oldmbox), buf, 4096)) && (rc == 0))
			{
			  errno = 0;
			  if ((chwrt = write (fileno (mbox), buf, lbuf)) < lbuf)
				{
				  syslog (LOG_INFO, "Error, too few bytes written on mailbox: %s", oldmailbox);
				  rc = ERR_FILE;
				}
			  if (errno)
				{
				  syslog (LOG_INFO, "Error %d on mailbox: %s", errno, strerror (errno));
				  rc = errno;
				}
			}
		}
/* unlink oldmailbox */

	  if (rc == 0)
		rc = unlink (oldmailbox);
	  if (rc != 0)
		syslog (LOG_ERR, "Failed unlink() :%s %s\n", oldmailbox, strerror (errno));
	}
  fl.l_type = F_UNLCK;
  fl.l_whence = SEEK_SET;
  fcntl (fileno (oldmbox), F_SETLK, &fl);

  fclose (oldmbox);
  fsync (fileno (mbox));
  rewind (mbox);
  free (oldmailbox);

  return rc;
}


/* Get current time. */
void
get_now (struct timeval *time, int debug)
{
  if (gettimeofday (time, NULL) != 0)
	{
	  fprintf (stderr, "Can't get current time.\n");
	}
  return;
}

/* Convert "struct timeval" to fractional seconds. */
double
time_to_seconds (struct timeval *tstart, struct timeval *tfinish)
{
  double t;

  t = (tfinish->tv_sec - tstart->tv_sec) + (tfinish->tv_usec - tstart->tv_usec) / 1e6;
  return t;
}
